@@ -2,14 +2,21 @@ use std::cell::RefCell;
 use std::fs;
 use std::path::{Path, PathBuf};
 use std::sync::mpsc::channel;
+use std::sync::Arc;
 use std::time::Duration;
 use std::time::Instant;
 
+#[cfg(not(target_arch = "wasm32"))]
+mod injection;
+
 use notify::{RecommendedWatcher, RecursiveMode, Watcher};
 
-use calcit_runner::{builtins, call_stack, cli_args, codegen, program, runner, snapshot, util};
+use calcit_runner::{
+  builtins, call_stack, cli_args, codegen, codegen::emit_js::gen_stack, codegen::COMPILE_ERRORS_FILE, program, runner, snapshot, util,
+};
+use im_ternary_tree::TernaryTreeList;
 
-struct ProgramSettings {
+pub struct ProgramSettings {
   entry_path: PathBuf,
   emit_path: String,
   reload_libs: bool,
@@ -17,10 +24,13 @@ struct ProgramSettings {
   emit_ir: bool,
 }
 
-pub const COMPILE_ERRORS_FILE: &str = "calcit.build-errors";
-
 fn main() -> Result<(), String> {
   builtins::effects::init_effects_states();
+
+  // get dirty functions injected
+  #[cfg(not(target_arch = "wasm32"))]
+  injection::inject_platform_apis();
+
   let cli_matches = cli_args::parse_cli();
   let settings = ProgramSettings {
     // has default value
@@ -43,10 +53,18 @@ fn main() -> Result<(), String> {
     eval_once = true;
     match snapshot::create_file_from_snippet(snippet) {
       Ok(main_file) => {
-        snapshot.files.insert(String::from("app.main"), main_file);
+        snapshot.files.insert(String::from("app.main").into(), main_file);
       }
       Err(e) => return Err(e),
     }
+    if let Some(cli_deps) = cli_matches.values_of("dep") {
+      for module_path in cli_deps {
+        let module_data = calcit_runner::load_module(module_path, settings.entry_path.parent().unwrap())?;
+        for (k, v) in &module_data.files {
+          snapshot.files.insert(k.to_owned(), v.to_owned());
+        }
+      }
+    }
   } else {
     // load entry file
     let content = fs::read_to_string(settings.entry_path.to_owned())
@@ -63,40 +81,46 @@ fn main() -> Result<(), String> {
       }
     }
   }
-  let init_fn = cli_matches
-    .value_of("init-fn")
-    .or(Some(&snapshot.configs.init_fn))
-    .unwrap();
-  let reload_fn = cli_matches
-    .value_of("reload-fn")
-    .or(Some(&snapshot.configs.reload_fn))
-    .unwrap();
+  let init_fn = cli_matches.value_of("init-fn").or(Some(&snapshot.configs.init_fn)).unwrap();
+  let reload_fn = cli_matches.value_of("reload-fn").or(Some(&snapshot.configs.reload_fn)).unwrap();
+
   // attach core
   for (k, v) in core_snapshot.files {
     snapshot.files.insert(k.to_owned(), v.to_owned());
   }
 
-  let mut program_code = program::extract_program_data(&snapshot)?;
+  // now global states
+  {
+    let mut prgm = { program::PROGRAM_CODE_DATA.write().unwrap() };
+    *prgm = program::extract_program_data(&snapshot)?;
+  }
+
   let check_warnings: &RefCell<Vec<String>> = &RefCell::new(vec![]);
 
   // make sure builtin classes are touched
   runner::preprocess::preprocess_ns_def(
-    calcit_runner::primes::CORE_NS,
-    calcit_runner::primes::BUILTIN_CLASSES_ENTRY,
-    &program_code,
-    calcit_runner::primes::BUILTIN_CLASSES_ENTRY,
+    calcit_runner::primes::CORE_NS.into(),
+    calcit_runner::primes::BUILTIN_CLASSES_ENTRY.into(),
+    calcit_runner::primes::BUILTIN_CLASSES_ENTRY.into(),
     None,
     check_warnings,
-  )?;
+    &rpds::List::new_sync(),
+  )
+  .map_err(|e| e.msg)?;
 
   let task = if settings.emit_js {
-    run_codegen(init_fn, reload_fn, &program_code, &settings.emit_path, false)
+    run_codegen(init_fn, reload_fn, &settings.emit_path, false)
   } else if settings.emit_ir {
-    run_codegen(init_fn, reload_fn, &program_code, &settings.emit_path, true)
+    run_codegen(init_fn, reload_fn, &settings.emit_path, true)
   } else {
     let started_time = Instant::now();
 
-    let v = calcit_runner::run_program(init_fn, im::vector![], &program_code)?;
+    let v = calcit_runner::run_program(init_fn, TernaryTreeList::Empty).map_err(|e| {
+      for w in e.warnings {
+        println!("{}", w);
+      }
+      e.msg
+    })?;
 
     let duration = Instant::now().duration_since(started_time);
     println!("took {}ms: {}", duration.as_micros() as f64 / 1000.0, v);
@@ -116,63 +140,65 @@ fn main() -> Result<(), String> {
   }
 
   if !eval_once {
-    println!("\nRunner: in watch mode...\n");
-    let (tx, rx) = channel();
-    let mut watcher: RecommendedWatcher = Watcher::new(tx, Duration::from_millis(200)).unwrap();
+    runner::track::track_task_add();
+    let copied_init_fn = Arc::new(init_fn.to_owned());
+    let copied_reload_fn = Arc::new(reload_fn.to_owned());
+    let copied_assets = Arc::new(assets_watch.map(|s| s.to_owned()).to_owned());
+    let copied_settings = Arc::new(settings);
+    std::thread::spawn(move || watch_files(copied_init_fn, copied_reload_fn, copied_settings, copied_assets));
+  }
+  runner::track::exit_when_cleared();
+  Ok(())
+}
 
-    let inc_path = settings.entry_path.parent().unwrap().join(".compact-inc.cirru");
-    if !inc_path.exists() {
-      fs::write(&inc_path, "").map_err(|e| -> String { e.to_string() })?;
-    }
+pub fn watch_files(init_fn: Arc<String>, reload_fn: Arc<String>, settings: Arc<ProgramSettings>, assets_watch: Arc<Option<String>>) {
+  println!("\nRunner: in watch mode...\n");
+  let (tx, rx) = channel();
+  let mut watcher: RecommendedWatcher = Watcher::new(tx, Duration::from_millis(200)).unwrap();
 
-    watcher.watch(&inc_path, RecursiveMode::NonRecursive).unwrap();
+  let inc_path = settings.entry_path.parent().unwrap().join(".compact-inc.cirru");
+  if !inc_path.exists() {
+    if let Err(e) = fs::write(&inc_path, "").map_err(|e| -> String { e.to_string() }) {
+      println!("file writing error: {}", e);
+    };
+  }
 
-    if let Some(assets_folder) = assets_watch {
-      watcher.watch(assets_folder, RecursiveMode::Recursive).unwrap();
-      println!("assets to watch: {}", assets_folder);
-    }
+  watcher.watch(&inc_path, RecursiveMode::NonRecursive).unwrap();
 
-    loop {
-      let mut change_happened = false;
-      match rx.recv() {
-        Ok(event) => {
-          match event {
-            notify::DebouncedEvent::NoticeWrite(..) => {
-              // ignored
+  if let Some(assets_folder) = assets_watch.as_ref() {
+    watcher.watch(assets_folder, RecursiveMode::Recursive).unwrap();
+    println!("assets to watch: {}", assets_folder);
+  };
+
+  loop {
+    match rx.recv() {
+      Ok(event) => {
+        match event {
+          notify::DebouncedEvent::Write(_) | notify::DebouncedEvent::Create(_) => {
+            // load new program code
+            let content = fs::read_to_string(&inc_path).expect("reading inc file");
+            if content.trim().is_empty() {
+              println!("failed re-compiling, got empty inc file");
+              continue;
             }
-            notify::DebouncedEvent::Write(_) => {
-              // mark state dirty
-              change_happened = true;
-            }
-            _ => println!("other file event: {:?}, ignored", event),
+            if let Err(e) = recall_program(&content, &init_fn, &reload_fn, &settings) {
+              println!("error: {}", e);
+            };
           }
+          // ignore other events
+          notify::DebouncedEvent::NoticeWrite(..) => {}
+          _ => println!("other file event: {:?}, ignored", event),
         }
-        Err(e) => println!("watch error: {:?}", e),
-      }
-      if change_happened {
-        // load new program code
-        let content = fs::read_to_string(&inc_path).unwrap();
-        if content.trim() == "" {
-          println!("failed re-compiling, got empty inc file");
-          continue;
-        }
-        recall_program(&mut program_code, &content, init_fn, reload_fn, &settings)?;
       }
+      Err(e) => println!("watch error: {:?}", e),
     }
-  } else {
-    Ok(())
   }
 }
 
-fn recall_program(
-  program_code: &mut program::ProgramCodeData,
-  content: &str,
-  init_fn: &str,
-  reload_fn: &str,
-  settings: &ProgramSettings,
-) -> Result<(), String> {
+// overwrite previous state
+
+fn recall_program(content: &str, init_fn: &str, reload_fn: &str, settings: &ProgramSettings) -> Result<(), String> {
   println!("\n-------- file change --------\n");
-  call_stack::clear_stack();
 
   // Steps:
   // 1. load changes file, and patch to program_code
@@ -184,7 +210,7 @@ fn recall_program(
 
   // println!("\ndata: {}", &data);
   // println!("\nchanges: {:?}", changes);
-  let new_code = program::apply_code_changes(program_code, &changes)?;
+  program::apply_code_changes(&changes)?;
   // println!("\nprogram code: {:?}", new_code);
 
   // clear data in evaled states
@@ -192,13 +218,18 @@ fn recall_program(
   builtins::meta::force_reset_gensym_index()?;
 
   let task = if settings.emit_js {
-    run_codegen(init_fn, reload_fn, &new_code, &settings.emit_path, false)
+    run_codegen(init_fn, reload_fn, &settings.emit_path, false)
   } else if settings.emit_ir {
-    run_codegen(init_fn, reload_fn, &new_code, &settings.emit_path, true)
+    run_codegen(init_fn, reload_fn, &settings.emit_path, true)
   } else {
     // run from `reload_fn` after reload
     let started_time = Instant::now();
-    let v = calcit_runner::run_program(reload_fn, im::vector![], &new_code)?;
+    let v = calcit_runner::run_program(reload_fn, TernaryTreeList::Empty).map_err(|e| {
+      for w in e.warnings {
+        println!("{}", w);
+      }
+      e.msg
+    })?;
     let duration = Instant::now().duration_since(started_time);
     println!("took {}ms: {}", duration.as_micros() as f64 / 1000.0, v);
     Ok(())
@@ -211,19 +242,10 @@ fn recall_program(
     }
   }
 
-  // overwrite previous state
-  *program_code = new_code;
-
   Ok(())
 }
 
-fn run_codegen(
-  init_fn: &str,
-  reload_fn: &str,
-  program_code: &program::ProgramCodeData,
-  emit_path: &str,
-  ir_mode: bool,
-) -> Result<(), String> {
+fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -> Result<(), String> {
   let started_time = Instant::now();
 
   let (init_ns, init_def) = util::string::extract_ns_def(init_fn)?;
@@ -243,33 +265,47 @@ fn run_codegen(
   let js_file_path = code_emit_path.join(format!("{}.js", COMPILE_ERRORS_FILE)); // TODO mjs_mode
 
   let check_warnings: &RefCell<Vec<String>> = &RefCell::new(vec![]);
+  gen_stack::clear_stack();
 
   // preprocess to init
-  match runner::preprocess::preprocess_ns_def(&init_ns, &init_def, program_code, &init_def, None, check_warnings) {
+  match runner::preprocess::preprocess_ns_def(
+    init_ns.to_owned().into(),
+    init_def.to_owned().into(),
+    init_def.to_owned().into(),
+    None,
+    check_warnings,
+    &rpds::List::new_sync(),
+  ) {
     Ok(_) => (),
     Err(failure) => {
       println!("\nfailed preprocessing, {}", failure);
-      call_stack::display_stack(&failure)?;
+      call_stack::display_stack(&failure.msg, &failure.stack)?;
 
       let _ = fs::write(
         &js_file_path,
         format!(
           "export default \"Preprocessing failed:\\n{}\";",
-          failure.trim().escape_default()
+          failure.msg.trim().escape_default()
         ),
       );
-
-      return Err(failure);
+      return Err(failure.msg);
     }
   }
 
   // preprocess to reload
-  match runner::preprocess::preprocess_ns_def(&reload_ns, &reload_def, program_code, &init_def, None, check_warnings) {
+  match runner::preprocess::preprocess_ns_def(
+    reload_ns.into(),
+    reload_def.into(),
+    init_def.to_owned().into(),
+    None,
+    check_warnings,
+    &rpds::List::new_sync(),
+  ) {
     Ok(_) => (),
     Err(failure) => {
       println!("\nfailed preprocessing, {}", failure);
-      call_stack::display_stack(&failure)?;
-      return Err(failure);
+      call_stack::display_stack(&failure.msg, &failure.stack)?;
+      return Err(failure.msg);
     }
   }
   let warnings = check_warnings.to_owned().into_inner();
@@ -280,10 +316,7 @@ fn run_codegen(
       content = format!("{}\n{}", content, message);
     }
 
-    let _ = fs::write(
-      &js_file_path,
-      format!("export default \"{}\";", content.trim().escape_default()),
-    );
+    let _ = fs::write(&js_file_path, format!("export default \"{}\";", content.trim().escape_default()));
     return Err(format!(
       "Found {} warnings, codegen blocked. errors in {}.js",
       warnings.len(),
@@ -302,17 +335,17 @@ fn run_codegen(
       Ok(_) => (),
       Err(failure) => {
         println!("\nfailed codegen, {}", failure);
-        call_stack::display_stack(&failure)?;
+        call_stack::display_stack(&failure, &gen_stack::get_gen_stack())?;
         return Err(failure);
       }
     }
   } else {
     // TODO entry ns
-    match codegen::emit_js::emit_js(&init_ns, emit_path) {
+    match codegen::emit_js::emit_js(&init_ns.to_owned(), emit_path) {
       Ok(_) => (),
       Err(failure) => {
         println!("\nfailed codegen, {}", failure);
-        call_stack::display_stack(&failure)?;
+        call_stack::display_stack(&failure, &gen_stack::get_gen_stack())?;
         return Err(failure);
       }
     }
