@@ -18,8 +18,6 @@ use std::cell::RefCell;
 use std::fs;
 use std::path::Path;
 
-use im_ternary_tree::TernaryTreeList;
-
 pub use primes::{Calcit, CalcitErr, CalcitItems};
 
 pub fn load_core_snapshot() -> Result<snapshot::Snapshot, String> {
@@ -37,12 +35,12 @@ pub fn run_program(init_fn: &str, params: CalcitItems) -> Result<Calcit, CalcitE
 
   // preprocess to init
   match runner::preprocess::preprocess_ns_def(
-    &init_ns.to_owned().into_boxed_str(),
-    &init_def.to_owned().into_boxed_str(),
-    &init_def.to_owned().into_boxed_str(),
+    init_ns.to_owned().into(),
+    init_def.to_owned().into(),
+    init_def.to_owned().into(),
     None,
     check_warnings,
-    &TernaryTreeList::Empty,
+    &rpds::List::new_sync(),
   ) {
     Ok(_) => (),
     Err(failure) => {
@@ -57,14 +55,16 @@ pub fn run_program(init_fn: &str, params: CalcitItems) -> Result<Calcit, CalcitE
     return Err(CalcitErr {
       msg: format!("Found {} warnings, runner blocked", warnings.len()),
       warnings: warnings.to_owned(),
-      stack: TernaryTreeList::Empty,
+      stack: rpds::List::new_sync(),
     });
   }
   match program::lookup_evaled_def(&init_ns, &init_def) {
     None => CalcitErr::err_str(format!("entry not initialized: {}/{}", init_ns, init_def)),
     Some(entry) => match entry {
-      Calcit::Fn(_, f_ns, _, def_scope, args, body) => {
-        let result = runner::run_fn(&params, &def_scope, &args, &body, &f_ns, &TernaryTreeList::Empty);
+      Calcit::Fn {
+        def_ns, scope, args, body, ..
+      } => {
+        let result = runner::run_fn(&params, &scope, &args, &body, def_ns, &rpds::List::new_sync());
         match result {
           Ok(v) => Ok(v),
           Err(failure) => {
