@@ -1,73 +1,95 @@
+mod syntax_name;
+
 use core::cmp::Ord;
-use regex::Regex;
 use std::cmp::Eq;
 use std::cmp::Ordering;
 use std::cmp::Ordering::*;
 use std::fmt;
 use std::hash::{Hash, Hasher};
+use std::sync::atomic::{AtomicUsize, Ordering::SeqCst};
+use std::sync::Arc;
 
-// String from nanoid!
-pub type NanoId = String;
+use cirru_edn::EdnKwd;
+use im_ternary_tree::TernaryTreeList;
+
+static ID_GEN: AtomicUsize = AtomicUsize::new(0);
 
 // scope
-pub type CalcitScope = im::HashMap<String, Calcit>;
-pub type CalcitItems = im::Vector<Calcit>;
+pub type CalcitScope = rpds::HashTrieMapSync<Arc<str>, Calcit>;
+pub type CalcitItems = TernaryTreeList<Calcit>;
+
+pub use syntax_name::CalcitSyntax;
+
+use crate::call_stack::CallStackList;
 
 #[derive(Debug, Clone, PartialEq)]
 pub enum SymbolResolved {
   ResolvedLocal,
-  ResolvedRaw,                                     // raw syntax, no target
-  ResolvedDef(String, String, Option<ImportRule>), // ns, def
+  ResolvedRaw, // raw syntax, no target
+  ResolvedDef {
+    ns: Arc<str>,
+    def: Arc<str>,
+    rule: Option<Arc<ImportRule>>,
+  }, // ns, def
 }
 
 /// defRule: ns def
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum ImportRule {
-  NsAs(String),               // ns
-  NsReferDef(String, String), // ns, def
-  NsDefault(String),          // ns, js only
+  NsAs(Arc<str>),                 // ns
+  NsReferDef(Arc<str>, Arc<str>), // ns, def
+  NsDefault(Arc<str>),            // ns, js only
 }
 
 /// special types wraps vector of calcit data for displaying
 #[derive(Debug, Clone, PartialEq, PartialOrd)]
-pub struct CrListWrap(pub im::Vector<Calcit>);
+pub struct CrListWrap(pub TernaryTreeList<Calcit>);
+
+pub struct CalcitList(pub Arc<TernaryTreeList<Calcit>>);
 
 #[derive(Debug, Clone)]
 pub enum Calcit {
   Nil,
   Bool(bool),
   Number(f64),
-  Symbol(String, String, String, Option<SymbolResolved>), // content, ns... so it has meta information
-  Keyword(String),
-  Str(String),
-  Thunk(Box<Calcit>, Option<Box<Calcit>>),
+  Symbol {
+    sym: Arc<str>,
+    ns: Arc<str>,
+    at_def: Arc<str>,
+    resolved: Option<Arc<SymbolResolved>>,
+  }, // content, ns... so it has meta information
+  Keyword(EdnKwd),
+  Str(Arc<str>),
+  Thunk(Arc<Calcit>, Option<Arc<Calcit>>),
   /// holding a path to its state
-  Ref(String),
+  Ref(Arc<str>),
   /// more tagged union type, more like an internal structure
-  Tuple(Box<Calcit>, Box<Calcit>),
+  Tuple(Arc<Calcit>, Arc<Calcit>),
+  ///  to be used by FFIs
+  Buffer(Vec<u8>),
   /// not for data, but for recursion
   Recur(CalcitItems),
   List(CalcitItems),
-  Set(im::HashSet<Calcit>),
-  Map(im::HashMap<Calcit, Calcit>),
-  Record(String, Vec<String>, Vec<Calcit>),
-  Proc(String),
-  Macro(
-    String, // name
-    String, // ns
-    NanoId,
-    CalcitItems, // args
-    CalcitItems, // body
-  ),
-  Fn(
-    String, // name
-    String, // ns
-    NanoId,
-    CalcitScope,
-    CalcitItems, // args
-    CalcitItems, // body
-  ),
-  Syntax(String, String), // name, ns... notice that `ns` is a meta info
+  Set(rpds::HashTrieSetSync<Calcit>),
+  Map(rpds::HashTrieMapSync<Calcit, Calcit>),
+  Record(EdnKwd, Arc<Vec<EdnKwd>>, Arc<Vec<Calcit>>), // usize of keyword id
+  Proc(Arc<str>),
+  Macro {
+    name: Arc<str>,         // name
+    def_ns: Arc<str>,       // ns
+    id: Arc<str>,           // an id
+    args: Arc<CalcitItems>, // args
+    body: Arc<CalcitItems>, // body
+  },
+  Fn {
+    name: Arc<str>,   // name
+    def_ns: Arc<str>, // ns
+    id: Arc<str>,     // an id
+    scope: CalcitScope,
+    args: Arc<CalcitItems>, // args
+    body: Arc<CalcitItems>, // body
+  },
+  Syntax(CalcitSyntax, Arc<str>), // name, ns... notice that `ns` is a meta info
 }
 
 impl fmt::Display for Calcit {
@@ -76,13 +98,10 @@ impl fmt::Display for Calcit {
       Calcit::Nil => f.write_str("nil"),
       Calcit::Bool(v) => f.write_str(&format!("{}", v)),
       Calcit::Number(n) => f.write_str(&format!("{}", n)),
-      Calcit::Symbol(s, ..) => f.write_str(&format!("'{}", s)),
-      Calcit::Keyword(s) => f.write_str(&format!(":{}", s)),
+      Calcit::Symbol { sym, .. } => f.write_str(&format!("'{}", sym)),
+      Calcit::Keyword(s) => f.write_str(&format!(":{}", s.to_string())),
       Calcit::Str(s) => {
-        lazy_static! {
-          static ref RE_SIMPLE_TOKEN: Regex = Regex::new(r"^[\w\d\-\?!\|]+$").unwrap();
-        }
-        if RE_SIMPLE_TOKEN.is_match(s) {
+        if is_simple_str(s) {
           write!(f, "|{}", s)
         } else {
           write!(f, "\"|{}\"", s.escape_default())
@@ -94,6 +113,29 @@ impl fmt::Display for Calcit {
       },
       Calcit::Ref(name) => f.write_str(&format!("(&ref {})", name)),
       Calcit::Tuple(a, b) => f.write_str(&format!("(:: {} {})", a, b)),
+      Calcit::Buffer(buf) => {
+        f.write_str("(&buffer")?;
+        if buf.len() > 8 {
+          f.write_str(&format!(
+            " {} {} {} {} {} {} {} {} ..+{}",
+            buffer_bit_hex(buf[0]),
+            buffer_bit_hex(buf[1]),
+            buffer_bit_hex(buf[2]),
+            buffer_bit_hex(buf[3]),
+            buffer_bit_hex(buf[4]),
+            buffer_bit_hex(buf[5]),
+            buffer_bit_hex(buf[6]),
+            buffer_bit_hex(buf[7]),
+            buf.len() - 8
+          ))?;
+        } else {
+          for b in buf {
+            f.write_str(" ")?;
+            f.write_str(&buffer_bit_hex(b.to_owned()))?;
+          }
+        }
+        f.write_str(")")
+      }
       Calcit::Recur(xs) => {
         f.write_str("(&recur")?;
         for x in xs {
@@ -124,17 +166,17 @@ impl fmt::Display for Calcit {
         Ok(())
       }
       Calcit::Record(name, fields, values) => {
-        f.write_str(&format!("(%{{}} {}", name))?;
+        f.write_str(&format!("(%{{}} {}", Calcit::Keyword(name.to_owned())))?;
         for idx in 0..fields.len() {
-          f.write_str(&format!(" ({} {})", fields[idx], values[idx]))?;
+          f.write_str(&format!(" ({} {})", Calcit::Keyword(fields[idx].to_owned()), values[idx]))?;
         }
         f.write_str(")")
       }
       Calcit::Proc(name) => f.write_str(&format!("(&proc {})", name)),
-      Calcit::Macro(name, _def_ns, _, args, body) => {
+      Calcit::Macro { name, args, body, .. } => {
         f.write_str(&format!("(&macro {} (", name))?;
         let mut need_space = false;
-        for a in args {
+        for a in &**args {
           if need_space {
             f.write_str(" ")?;
           }
@@ -143,7 +185,7 @@ impl fmt::Display for Calcit {
         }
         f.write_str(") (")?;
         need_space = false;
-        for b in body {
+        for b in &**body {
           if need_space {
             f.write_str(" ")?;
           }
@@ -152,10 +194,10 @@ impl fmt::Display for Calcit {
         }
         f.write_str("))")
       }
-      Calcit::Fn(name, _, _, _, args, body) => {
+      Calcit::Fn { name, args, body, .. } => {
         f.write_str(&format!("(&fn {} (", name))?;
         let mut need_space = false;
-        for a in args {
+        for a in &**args {
           if need_space {
             f.write_str(" ")?;
           }
@@ -164,7 +206,7 @@ impl fmt::Display for Calcit {
         }
         f.write_str(") (")?;
         need_space = false;
-        for b in body {
+        for b in &**body {
           if need_space {
             f.write_str(" ")?;
           }
@@ -177,8 +219,21 @@ impl fmt::Display for Calcit {
     }
   }
 }
-/// special types wraps vector of calcit data for displaying
 
+fn is_simple_str(tok: &str) -> bool {
+  for c in tok.chars() {
+    if !matches!(c, 'a'..='z' | 'A'..='Z' | '0'..='9' | '-' | '?' | '!' | '|') {
+      return false;
+    }
+  }
+  true
+}
+
+fn buffer_bit_hex(n: u8) -> String {
+  hex::encode(vec![n])
+}
+
+/// special types wraps vector of calcit data for displaying
 impl fmt::Display for CrListWrap {
   fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
     f.write_str(&format_to_lisp(&Calcit::List(self.0.to_owned()))) // TODO performance
@@ -190,7 +245,7 @@ pub fn format_to_lisp(x: &Calcit) -> String {
   match x {
     Calcit::List(ys) => {
       let mut s = String::from("(");
-      for (idx, y) in ys.iter().enumerate() {
+      for (idx, y) in ys.into_iter().enumerate() {
         if idx > 0 {
           s.push(' ');
         }
@@ -199,9 +254,9 @@ pub fn format_to_lisp(x: &Calcit) -> String {
       s.push(')');
       s
     }
-    Calcit::Symbol(s, ..) => s.to_owned(),
-    Calcit::Syntax(s, _ns) => s.to_owned(),
-    Calcit::Proc(s) => s.to_owned(),
+    Calcit::Symbol { sym, .. } => sym.to_string(),
+    Calcit::Syntax(s, _ns) => s.to_string(),
+    Calcit::Proc(s) => s.to_string(),
     a => format!("{}", a),
   }
 }
@@ -222,9 +277,9 @@ impl Hash for Calcit {
         // TODO https://stackoverflow.com/q/39638363/883571
         (*n as usize).hash(_state)
       }
-      Calcit::Symbol(s, _ns, _at_def, _resolved) => {
+      Calcit::Symbol { sym, .. } => {
         "symbol:".hash(_state);
-        s.hash(_state);
+        sym.hash(_state);
         // probaly no need, also won't be used in hashing
         // ns.hash(_state);
       }
@@ -249,6 +304,10 @@ impl Hash for Calcit {
         a.hash(_state);
         b.hash(_state);
       }
+      Calcit::Buffer(buf) => {
+        "buffer:".hash(_state);
+        buf.hash(_state);
+      }
       Calcit::Recur(v) => {
         "list:".hash(_state);
         v.hash(_state);
@@ -281,12 +340,12 @@ impl Hash for Calcit {
         "proc:".hash(_state);
         name.hash(_state);
       }
-      Calcit::Macro(_name, _ns, gen_id, ..) => {
+      Calcit::Macro { id: gen_id, .. } => {
         "macro:".hash(_state);
         // name.hash(_state);
         gen_id.hash(_state);
       }
-      Calcit::Fn(_name, _ns, gen_id, ..) => {
+      Calcit::Fn { id: gen_id, .. } => {
         "fn:".hash(_state);
         // name.hash(_state);
         gen_id.hash(_state);
@@ -294,7 +353,7 @@ impl Hash for Calcit {
       Calcit::Syntax(name, _ns) => {
         "syntax:".hash(_state);
         // syntax name can be used as identity
-        name.hash(_state);
+        name.to_string().hash(_state); // TODO
       }
     }
   }
@@ -323,9 +382,9 @@ impl Ord for Calcit {
       (Calcit::Number(_), _) => Less,
       (_, Calcit::Number(_)) => Greater,
 
-      (Calcit::Symbol(a, ..), Calcit::Symbol(b, ..)) => a.cmp(b),
-      (Calcit::Symbol(..), _) => Less,
-      (_, Calcit::Symbol(..)) => Greater,
+      (Calcit::Symbol { sym: a, .. }, Calcit::Symbol { sym: b, .. }) => a.cmp(b),
+      (Calcit::Symbol { .. }, _) => Less,
+      (_, Calcit::Symbol { .. }) => Greater,
 
       (Calcit::Keyword(a), Calcit::Keyword(b)) => a.cmp(b),
       (Calcit::Keyword(_), _) => Less,
@@ -350,6 +409,10 @@ impl Ord for Calcit {
       (Calcit::Tuple(_, _), _) => Less,
       (_, Calcit::Tuple(_, _)) => Greater,
 
+      (Calcit::Buffer(buf1), Calcit::Buffer(buf2)) => buf1.cmp(buf2),
+      (Calcit::Buffer(..), _) => Less,
+      (_, Calcit::Buffer(..)) => Greater,
+
       (Calcit::Recur(a), Calcit::Recur(b)) => a.cmp(b),
       (Calcit::Recur(_), _) => Less,
       (_, Calcit::Recur(_)) => Greater,
@@ -358,7 +421,7 @@ impl Ord for Calcit {
       (Calcit::List(_), _) => Less,
       (_, Calcit::List(_)) => Greater,
 
-      (Calcit::Set(a), Calcit::Set(b)) => match a.len().cmp(&b.len()) {
+      (Calcit::Set(a), Calcit::Set(b)) => match a.size().cmp(&b.size()) {
         Equal => {
           if a == b {
             Equal
@@ -388,13 +451,13 @@ impl Ord for Calcit {
       (Calcit::Proc(_), _) => Less,
       (_, Calcit::Proc(_)) => Greater,
 
-      (Calcit::Macro(_name1, _ns1, a, ..), Calcit::Macro(_name2, _ns2, b, ..)) => a.cmp(b),
-      (Calcit::Macro(..), _) => Less,
-      (_, Calcit::Macro(..)) => Greater,
+      (Calcit::Macro { id: a, .. }, Calcit::Macro { id: b, .. }) => a.cmp(b),
+      (Calcit::Macro { .. }, _) => Less,
+      (_, Calcit::Macro { .. }) => Greater,
 
-      (Calcit::Fn(_name1, _ns1, a, ..), Calcit::Fn(_name2, _ns2, b, ..)) => a.cmp(b), // compared with nanoid
-      (Calcit::Fn(..), _) => Less,
-      (_, Calcit::Fn(..)) => Greater,
+      (Calcit::Fn { id: a, .. }, Calcit::Fn { id: b, .. }) => a.cmp(b), // compared with nanoid
+      (Calcit::Fn { .. }, _) => Less,
+      (_, Calcit::Fn { .. }) => Greater,
 
       (Calcit::Syntax(a, _), Calcit::Syntax(b, _)) => a.cmp(b),
     }
@@ -415,12 +478,13 @@ impl PartialEq for Calcit {
       (Calcit::Nil, Calcit::Nil) => true,
       (Calcit::Bool(a), Calcit::Bool(b)) => a == b,
       (Calcit::Number(a), Calcit::Number(b)) => a == b,
-      (Calcit::Symbol(a, ..), Calcit::Symbol(b, ..)) => a == b,
+      (Calcit::Symbol { sym: a, .. }, Calcit::Symbol { sym: b, .. }) => a == b,
       (Calcit::Keyword(a), Calcit::Keyword(b)) => a == b,
       (Calcit::Str(a), Calcit::Str(b)) => a == b,
       (Calcit::Thunk(a, _), Calcit::Thunk(b, _)) => a == b,
       (Calcit::Ref(a), Calcit::Ref(b)) => a == b,
       (Calcit::Tuple(a, b), Calcit::Tuple(c, d)) => a == c && b == d,
+      (Calcit::Buffer(b), Calcit::Buffer(d)) => b == d,
       (Calcit::List(a), Calcit::List(b)) => a == b,
       (Calcit::Set(a), Calcit::Set(b)) => a == b,
       (Calcit::Map(a), Calcit::Map(b)) => a == b,
@@ -430,8 +494,8 @@ impl PartialEq for Calcit {
 
       // functions compared with nanoid
       (Calcit::Proc(a), Calcit::Proc(b)) => a == b,
-      (Calcit::Macro(_name1, _ns1, a, ..), Calcit::Macro(_name2, _ns2, b, ..)) => a == b,
-      (Calcit::Fn(_name1, _ns1, a, ..), Calcit::Fn(_name2, _ns2, b, ..)) => a == b,
+      (Calcit::Macro { id: a, .. }, Calcit::Macro { id: b, .. }) => a == b,
+      (Calcit::Fn { id: a, .. }, Calcit::Fn { id: b, .. }) => a == b,
       (Calcit::Syntax(a, _), Calcit::Syntax(b, _)) => a == b,
       (_, _) => false,
     }
@@ -443,11 +507,18 @@ pub const BUILTIN_CLASSES_ENTRY: &str = "&init-builtin-classes!";
 pub const GENERATED_NS: &str = "calcit.gen";
 pub const GENERATED_DEF: &str = "gen%";
 
+lazy_static! {
+  pub static ref GEN_NS: Arc<str> = GENERATED_NS.to_string().into();
+  pub static ref GEN_DEF: Arc<str> = GENERATED_DEF.to_string().into();
+  pub static ref GEN_CORE_NS: Arc<str> = CORE_NS.to_string().into();
+  pub static ref GEN_CLASS_ENTRY: Arc<str> = BUILTIN_CLASSES_ENTRY.to_string().into();
+}
+
 impl Calcit {
   pub fn turn_string(&self) -> String {
     match self {
       Calcit::Nil => String::from(""),
-      Calcit::Str(s) => s.to_owned(),
+      Calcit::Str(s) => (**s).to_owned(),
       _ => format!("{}", self),
     }
   }
@@ -455,4 +526,74 @@ impl Calcit {
   pub fn lisp_str(&self) -> String {
     format_to_lisp(self)
   }
+
+  pub fn new_str<T: Into<String>>(s: T) -> Calcit {
+    Calcit::Str(s.into().into())
+  }
+
+  /// makes sure that keyword is from global dict, not created by fresh
+  pub fn kwd(s: &str) -> Self {
+    Calcit::Keyword(EdnKwd::from(s))
+  }
+}
+
+/// too naive id generator to be safe in WASM
+pub fn gen_core_id() -> Arc<str> {
+  let c = ID_GEN.fetch_add(1, SeqCst);
+  format!("gen_id_{}", c).into()
+}
+
+#[derive(Debug, Clone, PartialEq)]
+pub struct CalcitErr {
+  pub msg: String,
+  pub warnings: Vec<String>,
+  pub stack: rpds::ListSync<crate::call_stack::CalcitStack>,
+}
+
+impl fmt::Display for CalcitErr {
+  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+    f.write_str(&self.msg)?;
+    if !self.warnings.is_empty() {
+      f.write_str("\n")?;
+      for w in &self.warnings {
+        writeln!(f, "{}", w)?;
+      }
+    }
+    Ok(())
+  }
+}
+
+impl From<String> for CalcitErr {
+  /// hope this does not add extra costs
+  fn from(msg: String) -> Self {
+    CalcitErr {
+      msg,
+      warnings: vec![],
+      stack: rpds::List::new_sync(),
+    }
+  }
+}
+
+impl CalcitErr {
+  pub fn use_str<T: Into<String>>(msg: T) -> Self {
+    CalcitErr {
+      msg: msg.into(),
+      warnings: vec![],
+      stack: rpds::List::new_sync(),
+    }
+  }
+  pub fn err_str<T: Into<String>>(msg: T) -> Result<Calcit, Self> {
+    Err(CalcitErr {
+      msg: msg.into(),
+      warnings: vec![],
+      stack: rpds::List::new_sync(),
+    })
+  }
+  pub fn use_msg_stack<T: Into<String>>(msg: T, stack: &CallStackList) -> Self {
+    CalcitErr {
+      msg: msg.into(),
+      warnings: vec![],
+      stack: stack.to_owned(),
+    }
+  }
 }
