@@ -1,3 +1,5 @@
+use std::sync::Arc;
+
 pub fn is_digit(c: char) -> bool {
   let n = c as u32;
   // ascii table https://tool.oschina.net/commons?type=4
@@ -77,17 +79,16 @@ pub fn matches_digits(xs: &str) -> bool {
 
 pub fn matches_js_var(xs: &str) -> bool {
   if xs.is_empty() {
-    false
-  } else {
-    for (idx, x) in xs.chars().enumerate() {
-      if is_letter(x) || x == '_' || x == '$' || (idx > 0 && is_digit(x)) {
-        // ok
-      } else {
-        return false;
-      }
-    }
-    true
+    return false;
   }
+  for (idx, x) in xs.chars().enumerate() {
+    if is_letter(x) || x == '_' || x == '$' || (idx > 0 && is_digit(x)) {
+      // ok
+    } else {
+      return false;
+    }
+  }
+  true
 }
 
 pub fn has_ns_part(x: &str) -> bool {
@@ -99,7 +100,7 @@ pub fn has_ns_part(x: &str) -> bool {
 
 /// js/JSON.stringify -like API
 pub fn wrap_js_str(s: &str) -> String {
-  let mut c: String = String::from("");
+  let mut c: String = String::with_capacity(s.len() + 2);
   c.push('"');
   for i in s.escape_default() {
     c.push(i);
@@ -109,7 +110,7 @@ pub fn wrap_js_str(s: &str) -> String {
 }
 
 pub fn extract_ns_def(s: &str) -> Result<(String, String), String> {
-  let pieces: Vec<&str> = (&s).split('/').collect();
+  let pieces: Vec<&str> = s.split('/').collect();
   if pieces.len() == 2 {
     Ok((pieces[0].to_owned(), pieces[1].to_owned()))
   } else {
@@ -117,12 +118,12 @@ pub fn extract_ns_def(s: &str) -> Result<(String, String), String> {
   }
 }
 
-pub fn extract_pkg_from_def(s: &str) -> Option<String> {
-  let pieces: Vec<&str> = (&s).split('/').collect();
+pub fn extract_pkg_from_def(s: &str) -> Option<Arc<str>> {
+  let pieces: Vec<&str> = s.split('/').collect();
   if !pieces.is_empty() {
-    let p2: Vec<&str> = (pieces[0]).split('.').collect();
+    let p2: Vec<&str> = pieces[0].split('.').collect();
     if !p2.is_empty() {
-      Some(p2[0].to_string())
+      Some(p2[0].to_owned().into())
     } else {
       None
     }
