@@ -2,9 +2,9 @@ import { overwriteComparator, initTernaryTreeMap } from "@calcit/ternary-tree";
 import { CirruWriterNode, writeCirruCode } from "@cirru/writer.ts";
 
 import { CalcitValue } from "./js-primes";
-import { CalcitList } from "./js-list";
+import { CalcitList, CalcitSliceList } from "./js-list";
 import { CalcitRecord } from "./js-record";
-import { CalcitMap } from "./js-map";
+import { CalcitMap, CalcitSliceMap } from "./js-map";
 import { CalcitSet } from "./js-set";
 import { CalcitKeyword, CalcitSymbol, CalcitRecur, CalcitRef, kwd } from "./calcit-data";
 import { CalcitTuple } from "./js-tuple";
@@ -44,11 +44,11 @@ export let to_cirru_edn = (x: CalcitValue): CirruEdnFormat => {
   if (x instanceof CalcitSymbol) {
     return x.toString();
   }
-  if (x instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
     // TODO can be faster
     return (["[]"] as CirruEdnFormat[]).concat(x.toArray().map(to_cirru_edn));
   }
-  if (x instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
     let buffer: CirruEdnFormat = ["{}"];
     for (let [k, v] of x.pairs()) {
       buffer.push([to_cirru_edn(k), to_cirru_edn(v)]);
@@ -56,10 +56,9 @@ export let to_cirru_edn = (x: CalcitValue): CirruEdnFormat => {
     return buffer;
   }
   if (x instanceof CalcitRecord) {
-    let result: Record<string, CalcitValue> = {};
-    let buffer: CirruEdnFormat = ["%{}", x.name];
-    for (let idx in x.fields) {
-      buffer.push([x.fields[idx], to_cirru_edn(x.values[idx])]);
+    let buffer: CirruEdnFormat = ["%{}", x.name.toString()];
+    for (let idx = 0; idx < x.fields.length; idx++) {
+      buffer.push([x.fields[idx].toString(), to_cirru_edn(x.values[idx])]);
     }
     return buffer;
   }
@@ -76,6 +75,8 @@ export let to_cirru_edn = (x: CalcitValue): CirruEdnFormat => {
     if (x.fst instanceof CalcitSymbol && x.fst.value === "quote") {
       // turn `x.snd` with CalcitList into raw Cirru nodes, which is in plain Array
       return ["quote", toWriterNode(x.snd as any)] as CirruEdnFormat;
+    } else if (x.fst instanceof CalcitRecord) {
+      return ["::", x.fst.name.toString(), to_cirru_edn(x.snd)];
     } else {
       throw new Error(`Unsupported tag for EDN: ${x.fst}`);
     }
@@ -84,6 +85,15 @@ export let to_cirru_edn = (x: CalcitValue): CirruEdnFormat => {
   throw new Error("Unexpected data to to-cirru-edn");
 };
 
+/** makes sure we got string */
+let extractFieldKwd = (x: string) => {
+  if (x[0] === ":") {
+    return kwd(x.slice(1));
+  } else {
+    return kwd(x);
+  }
+};
+
 export let extract_cirru_edn = (x: CirruEdnFormat): CalcitValue => {
   if (typeof x === "string") {
     if (x === "nil") {
@@ -95,7 +105,7 @@ export let extract_cirru_edn = (x: CirruEdnFormat): CalcitValue => {
     if (x === "false") {
       return false;
     }
-    if (x == "") {
+    if (x === "") {
       throw new Error("cannot be empty");
     }
     if (x[0] === "|" || x[0] === '"') {
@@ -119,46 +129,55 @@ export let extract_cirru_edn = (x: CirruEdnFormat): CalcitValue => {
       throw new Error("Cannot be empty");
     }
     if (x[0] === "{}") {
-      let result: Array<[CalcitValue, CalcitValue]> = [];
+      let result: Array<CalcitValue> = [];
       x.forEach((pair, idx) => {
-        if (idx == 0) {
+        if (idx === 0) {
           return; // skip first `{}` symbol
         }
-        if (pair instanceof Array && pair.length == 2) {
-          result.push([extract_cirru_edn(pair[0]), extract_cirru_edn(pair[1])]);
+        if (pair instanceof Array && pair.length === 2) {
+          result.push(extract_cirru_edn(pair[0]), extract_cirru_edn(pair[1]));
         } else {
           throw new Error("Expected pairs for map");
         }
       });
-      return new CalcitMap(initTernaryTreeMap(result));
+      return new CalcitSliceMap(result);
     }
     if (x[0] === "%{}") {
       let name = x[1];
       if (typeof name != "string") {
         throw new Error("Expected string for record name");
       }
-      let fields: Array<string> = [];
-      let values: Array<CalcitValue> = [];
+      // put to entries first, sort and then...
+      let entries: Array<[CalcitKeyword, CalcitValue]> = [];
       x.forEach((pair, idx) => {
         if (idx <= 1) {
           return; // skip %{} name
         }
 
-        if (pair instanceof Array && pair.length == 2) {
+        if (pair instanceof Array && pair.length === 2) {
           if (typeof pair[0] === "string") {
-            fields.push(pair[0]);
+            entries.push([extractFieldKwd(pair[0]), extract_cirru_edn(pair[1])]);
           } else {
             throw new Error("Expected string as field");
           }
-          values.push(extract_cirru_edn(pair[1]));
         } else {
           throw new Error("Expected pairs for map");
         }
       });
-      return new CalcitRecord(name, fields, values);
+      entries.sort((a, b) => {
+        return a[0].cmp(b[0]);
+      });
+      let fields: Array<CalcitKeyword> = [];
+      let values: Array<CalcitValue> = [];
+
+      for (let idx = 0; idx < entries.length; idx++) {
+        fields.push(entries[idx][0]);
+        values.push(entries[idx][1]);
+      }
+      return new CalcitRecord(extractFieldKwd(name), fields, values);
     }
     if (x[0] === "[]") {
-      return new CalcitList(x.slice(1).map(extract_cirru_edn));
+      return new CalcitSliceList(x.slice(1).map(extract_cirru_edn));
     }
     if (x[0] === "#{}") {
       return new CalcitSet(x.slice(1).map(extract_cirru_edn));
@@ -172,6 +191,12 @@ export let extract_cirru_edn = (x: CirruEdnFormat): CalcitValue => {
       }
       return new CalcitTuple(new CalcitSymbol("quote"), to_calcit_data(x[1], true));
     }
+    if (x[0] === "::") {
+      if (x.length !== 3) {
+        throw new Error("tuple expects 2 values");
+      }
+      return new CalcitTuple(extract_cirru_edn(x[1]), extract_cirru_edn(x[2]));
+    }
   }
   console.error(x);
   throw new Error("Unexpected data from cirru-edn");
@@ -184,10 +209,10 @@ export let format_cirru_edn = (data: CalcitValue, useInline: boolean = true): st
   if (typeof data === "string") {
     return "\ndo " + to_cirru_edn(data) + "\n";
   }
-  if (typeof data == "boolean") {
+  if (typeof data === "boolean") {
     return "\ndo " + to_cirru_edn(data) + "\n";
   }
-  if (typeof data == "string") {
+  if (typeof data === "string") {
     return "\ndo " + to_cirru_edn(data) + "\n";
   }
   if (data instanceof CalcitSymbol) {
@@ -219,7 +244,7 @@ export let to_calcit_data = (x: any, noKeyword: boolean = false): CalcitValue =>
     x.forEach((v) => {
       result.push(to_calcit_data(v, noKeyword));
     });
-    return new CalcitList(result);
+    return new CalcitSliceList(result);
   }
   if (x instanceof Set) {
     let result: Array<CalcitValue> = [];
@@ -229,8 +254,8 @@ export let to_calcit_data = (x: any, noKeyword: boolean = false): CalcitValue =>
     return new CalcitSet(result);
   }
 
-  if (x instanceof CalcitList) return x;
-  if (x instanceof CalcitMap) return x;
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) return x;
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) return x;
   if (x instanceof CalcitSet) return x;
   if (x instanceof CalcitRecord) return x;
   if (x instanceof CalcitRecur) return x;
@@ -241,22 +266,22 @@ export let to_calcit_data = (x: any, noKeyword: boolean = false): CalcitValue =>
 
   // detects object
   if (x === Object(x)) {
-    let result: Array<[CalcitValue, CalcitValue]> = [];
+    let result: Array<CalcitValue> = [];
     Object.keys(x).forEach((k) => {
-      result.push([to_calcit_data(k, noKeyword), to_calcit_data(x[k], noKeyword)]);
+      result.push(to_calcit_data(k, noKeyword), to_calcit_data(x[k], noKeyword));
     });
-    return new CalcitMap(initTernaryTreeMap(result));
+    return new CalcitSliceMap(result);
   }
 
   console.error(x);
   throw new Error("Unexpected data for converting");
 };
 
-let toWriterNode = (xs: CalcitList): CirruWriterNode => {
+let toWriterNode = (xs: CalcitList | CalcitSliceList): CirruWriterNode => {
   if (typeof xs === "string") {
     return xs;
   }
-  if (xs instanceof CalcitList) {
+  if (xs instanceof CalcitList || xs instanceof CalcitSliceList) {
     return xs.toArray().map(toWriterNode);
   } else {
     throw new Error("Unexpected type for CirruWriteNode");
